#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  4 15:25:33 2022

@author: tanu
"""
#%%
import os, sys
import pandas as pd
import numpy as np
import pprint as pp
from copy import deepcopy
from sklearn import linear_model
from sklearn import datasets
from collections import Counter

from sklearn.linear_model import LogisticRegression, LogisticRegressionCV
from sklearn.linear_model import RidgeClassifier, RidgeClassifierCV, SGDClassifier, PassiveAggressiveClassifier

from sklearn.naive_bayes import BernoulliNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier, ExtraTreeClassifier
from sklearn.ensemble import RandomForestClassifier, ExtraTreesClassifier, AdaBoostClassifier, GradientBoostingClassifier, BaggingClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.gaussian_process import GaussianProcessClassifier, kernels
from sklearn.gaussian_process.kernels import RBF, DotProduct, Matern, RationalQuadratic, WhiteKernel

from sklearn.discriminant_analysis import LinearDiscriminantAnalysis, QuadraticDiscriminantAnalysis
from sklearn.neural_network import MLPClassifier

from sklearn.svm import SVC
from xgboost import XGBClassifier
from sklearn.naive_bayes import MultinomialNB
from sklearn.preprocessing import StandardScaler, MinMaxScaler, OneHotEncoder

from sklearn.compose import ColumnTransformer
from sklearn.compose import make_column_transformer

from sklearn.metrics import make_scorer, confusion_matrix, accuracy_score, balanced_accuracy_score, precision_score, average_precision_score, recall_score
from sklearn.metrics import roc_auc_score, roc_curve, f1_score, matthews_corrcoef, jaccard_score, classification_report

from sklearn.model_selection import train_test_split, cross_validate, cross_val_score
from sklearn.model_selection import StratifiedKFold,RepeatedStratifiedKFold, RepeatedKFold

from sklearn.pipeline import Pipeline, make_pipeline

from sklearn.feature_selection import RFE, RFECV

import itertools
import seaborn as sns
import matplotlib.pyplot as plt

from statistics import mean, stdev, median, mode

from imblearn.over_sampling import RandomOverSampler
from imblearn.under_sampling import RandomUnderSampler
from imblearn.over_sampling import SMOTE
from sklearn.datasets import make_classification
from imblearn.combine import SMOTEENN
from imblearn.combine import SMOTETomek

from imblearn.over_sampling import SMOTENC
from imblearn.under_sampling import EditedNearestNeighbours
from imblearn.under_sampling import RepeatedEditedNearestNeighbours

from sklearn.model_selection import GridSearchCV
from sklearn.base import BaseEstimator
from sklearn.impute import KNNImputer as KNN
import json

#%% GLOBALS
rs = {'random_state': 42}
njobs = {'n_jobs': 10}

scoring_fn =  ({ 'mcc'        : make_scorer(matthews_corrcoef)
                , 'accuracy'      : make_scorer(accuracy_score)
                , 'fscore'     : make_scorer(f1_score)
                , 'precision'  : make_scorer(precision_score)
                , 'recall'     : make_scorer(recall_score)
                , 'roc_auc'    : make_scorer(roc_auc_score)
                , 'jcc'        : make_scorer(jaccard_score)
            }) 
  
skf_cv = StratifiedKFold(n_splits = 10
                          #, shuffle = False, random_state= None)
                           , shuffle = True,**rs)

rskf_cv = RepeatedStratifiedKFold(n_splits = 10
                                  , n_repeats = 3
                                  , **rs)

mcc_score_fn = {'mcc': make_scorer(matthews_corrcoef)}
jacc_score_fn = {'jcc': make_scorer(jaccard_score)}
#%%
# Multiple Classification - Model Pipeline
def MultModelsCl(input_df, target, skf_cv
                       , blind_test_input_df
                       , blind_test_target
                       , var_type = ['numerical', 'categorical','mixed']):

    '''
    @ param input_df: input features 
    @ type: df with input features WITHOUT the target variable
    
    @param target: target (or output) feature
    @type: df or np.array or Series
    
    @param skv_cv: stratifiedK fold int or object to allow shuffle and random state to pass
    @type: int or StratifiedKfold()
    
    @var_type: numerical, categorical and mixed to determine what col_transform to apply (MinMaxScalar and/or one-ho    t encoder)
    @type: list

    returns
    Dict containing multiple classification scores for each model and mean of each Stratified Kfold including training
    '''
    
    # Determine categorical and numerical features
    numerical_ix = input_df.select_dtypes(include=['int64', 'float64']).columns
    numerical_ix
    categorical_ix = input_df.select_dtypes(include=['object', 'bool']).columns
    categorical_ix    

    # Determine preprocessing steps ~ var_type
    if var_type == 'numerical':
        t = [('num', MinMaxScaler(), numerical_ix)]

    if var_type == 'categorical':
        t = [('cat', OneHotEncoder(), categorical_ix)]
    
    if var_type == 'mixed':
        t = [('num', MinMaxScaler(), numerical_ix)
            , ('cat', OneHotEncoder(), categorical_ix) ]
        
    col_transform = ColumnTransformer(transformers = t
                                       , remainder='passthrough')
    
    # Specify multiple Classification models
    lr      = LogisticRegression(**rs)
    lrcv    = LogisticRegressionCV(**rs)
    gnb     = GaussianNB()
    nb      = BernoulliNB()
    knn     = KNeighborsClassifier()
    svc     = SVC(**rs)
    mlp     = MLPClassifier(max_iter = 500, **rs)
    dt      = DecisionTreeClassifier(**rs)
    ets     = ExtraTreesClassifier(**rs)
    
    rf      = RandomForestClassifier(**rs, n_estimators = 1000 )
    rf2     = RandomForestClassifier(
                          min_samples_leaf = 5
                          , n_estimators     = 1000
                          , bootstrap        = True
                          , oob_score        = True
                          , **njobs
                          , **rs
                          , max_features     = 'auto')
    xgb = XGBClassifier(**rs, verbosity = 0, use_label_encoder =False)
                        
    lda = LinearDiscriminantAnalysis()
    
    mnb = MultinomialNB()
    
    pa  = PassiveAggressiveClassifier(**rs, **njobs)
    
    sgd = SGDClassifier(**rs, **njobs)  

    abc = AdaBoostClassifier(**rs)
    bc  = BaggingClassifier(**rs, **njobs, bootstrap = True, oob_score = True)
    et  = ExtraTreeClassifier(**rs)
    gpc = GaussianProcessClassifier(**rs)
    gbc = GradientBoostingClassifier(**rs)
    qda = QuadraticDiscriminantAnalysis()
    rc  = RidgeClassifier(**rs)
    rccv  = RidgeClassifierCV(cv = 10)
    
    models = [('Logistic Regression'       , lr)
            , ('Logistic RegressionCV'     , lrcv)
            , ('Gaussian NB'               , gnb)
            , ('Naive Bayes'               , nb)
            , ('K-Nearest Neighbors'       , knn) 
            , ('SVM'                       , svc) 
            , ('MLP'                       , mlp) 
            , ('Decision Tree'             , dt) 
            , ('Extra Trees'               , ets) 
            , ('Extra Tree'                , et)
            , ('Random Forest'             , rf) 
            , ('Random Forest2'            , rf2) 
            , ('Naive Bayes'               , nb)
            , ('XGBoost'                   , xgb)
            , ('LDA'                       , lda)
            , ('Multinomial'               , mnb)
            , ('Passive Aggresive'         , pa)
            , ('Stochastic GDescent'       , sgd)
            , ('AdaBoost Classifier'       , abc)
            , ('Bagging Classifier'        , bc)
            , ('Gaussian Process'          , gpc)
            , ('Gradient Boosting'         , gbc)
            , ('QDA'                       , qda)
            , ('Ridge Classifier'          , rc)
            , ('Ridge ClassifierCV'        , rccv)
            ]
        
    mm_skf_scoresD = {}
     
    for model_name, model_fn in models:
        print('\nModel_name:', model_name
        , '\nModel func:'    , model_fn
        , '\nList of models:', models)
    
        model_pipeline = Pipeline([
            ('prep'     , col_transform)
            , ('model'  , model_fn)])
            
        print('Running model pipeline:', model_pipeline)
        skf_cv_mod = cross_validate(model_pipeline
                              , input_df
                              , target
                              , cv = skf_cv
                              , scoring = scoring_fn
                              , return_train_score = True)
        mm_skf_scoresD[model_name] = {}
        for key, value in skf_cv_mod.items():
            print('\nkey:', key, '\nvalue:', value)
            print('\nmean value:', mean(value))
            mm_skf_scoresD[model_name][key] = round(mean(value),2)
            #pp.pprint(mm_skf_scoresD)
            #cvtrain_mcc = mm_skf_scoresD[model_name]['test_mcc']
            
    #return(mm_skf_scoresD)
#%%
        #=========================
        # Blind test: BTS results
        #=========================
        # Build the final results with all scores for a feature selected model
        #bts_predict = gscv_fs.predict(blind_test_input_df)
        model_pipeline.fit(input_df, target)
        bts_predict = model_pipeline.predict(blind_test_input_df)
        
        bts_mcc_score = round(matthews_corrcoef(blind_test_target, bts_predict),2)
        print('\nMCC on Blind test:'     , bts_mcc_score)
        print('\nAccuracy on Blind test:', round(accuracy_score(blind_test_target, bts_predict),2))
        
        # Diff b/w train and bts test scores
        #train_test_diff_MCC = cvtrain_mcc - bts_mcc_score
        # print('\nDiff b/w train and blind test score (MCC):', train_test_diff)
        
       
        # # create a dict with all scores
        # lr_btsD = { 'model_name': model_name
        #            , 'bts_mcc':None
        #                , 'bts_fscore':None
        #                , 'bts_precision':None
        #                , 'bts_recall':None
        #                , 'bts_accuracy':None
        #                , 'bts_roc_auc':None
        #                , 'bts_jaccard':None}
        
        
        mm_skf_scoresD[model_name]['bts_mcc']       = bts_mcc_score
        mm_skf_scoresD[model_name]['bts_fscore']    = round(f1_score(blind_test_target, bts_predict),2)
        mm_skf_scoresD[model_name]['bts_precision'] = round(precision_score(blind_test_target, bts_predict),2)
        mm_skf_scoresD[model_name]['bts_recall']    = round(recall_score(blind_test_target, bts_predict),2)
        mm_skf_scoresD[model_name]['bts_accuracy']  = round(accuracy_score(blind_test_target, bts_predict),2)
        mm_skf_scoresD[model_name]['bts_roc_auc']   = round(roc_auc_score(blind_test_target, bts_predict),2)
        mm_skf_scoresD[model_name]['bts_jaccard']   = round(jaccard_score(blind_test_target, bts_predict),2)
        #mm_skf_scoresD[model_name]['diff_mcc']      = train_test_diff_MCC

    return(mm_skf_scoresD)
    
